/*-------------------------------------------------
 *    Kernel version 2.4.20 - 2.6.8, 2.6.13
 *    RS485 (LPT-port)
 * 
 *    Andre Zaitsev, Olga Voziakova
 *    ovoz@sai.msu.ru
 *    $History:
 *
 *---------------------------------------- */
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kdev_t.h>
#include <linux/fs.h>
#include <linux/device.h>

#include <linux/errno.h>
#include <linux/slab.h>
#include <asm/uaccess.h>   //  copy_from_user
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <asm/irq.h>
#include <asm/io.h>
#include <linux/parport.h>
#include <linux/spinlock.h>

#include <mcua.h> // To define KERN_2_6
#ifdef KERN_2_6
   #include <linux/cdev.h>
   #include <linux/pnp.h>
#endif
#ifndef KERN_2_6_18
   #include <linux/devfs_fs_kernel.h>
#endif

#include <lpt.h>
#include <rs485.h>
#include <rs485_tools.h>
#include <rs485_irq.h>
#include <rserrno.h>
#include <rs485_fifo.h> 

static unsigned int major;
RS485INFO lpt_table[PARPORT_MAX];
#ifndef __PP_PC__
static int __initdata io[PARPORT_MAX] = {_IO_LPT1, [1 ... PARPORT_MAX-1]=_IO_LPT2 };
static int __initdata irq[PARPORT_MAX] = {_IRQ_LPT1, [1 ... PARPORT_MAX-1]=_IRQ_LPT2};
#endif
static int __initdata baudrate[PARPORT_MAX] = {_BAUDRATE_1, [1 ... PARPORT_MAX-1]=_BAUDRATE_2};


struct file_operations rs485_fops;
static volatile int data_ready = 0;
DECLARE_WAIT_QUEUE_HEAD(tr_queue);

#ifdef KERN_2_6
#ifndef __PP_PC__

static const struct pnp_device_id rs_lpt_pnp_tbl[] = {
        {.id = "PNP0400", .driver_data = 0},        // Standard LPT Printer Port
        {.id = "PNP0401", .driver_data = 0},        // ECP Printer Port
        { }
};

MODULE_DEVICE_TABLE(pnp, rs_lpt_pnp_tbl);

static int rs_lpt_pnp_probe(struct pnp_dev *dev, const struct pnp_device_id *id)
{
    return 0;
}

static void rs_lpt_pnp_remove(struct pnp_dev *dev)
{ }

static struct pnp_driver rs_lpt_pnp_driver = {
    .name           = "parport_rs485",
    .id_table       = rs_lpt_pnp_tbl,
    .probe          = rs_lpt_pnp_probe,
    .remove         = rs_lpt_pnp_remove,
};

static int pnp_registered_parport;

void rs485_activate_parport(void) {
    int err;
    err = pnp_register_driver(&rs_lpt_pnp_driver);
    if(!err) pnp_registered_parport = 1;
}

void rs485_stop_parport(void) {
    if(pnp_registered_parport)
        pnp_unregister_driver(&rs_lpt_pnp_driver);
}
#endif
#endif

/* Don't ask why the stuff is implemented in the way it is.

rs_lock is used here to protect ONLY from dereferring NULL in interrupt context.
sem is used to protect consecutive open and release calls. */
inline int rs485_init_rs(RS485INFO* pinfo) {
	int ret = 0;
	unsigned long irqflags;

	spin_lock_irqsave(&pinfo->rs_lock, irqflags);

	if (pinfo->use_mcua == 0) {
		pinfo->rs = (struct RS *)kmalloc(sizeof(struct RS), GFP_ATOMIC);
		if (!pinfo->rs) {
			ret = -ENOMEM;
			goto end;
		}
	}
	ret = pinfo->use_mcua++;
end:
	spin_unlock_irqrestore(&pinfo->rs_lock, irqflags);
	return ret;
}
inline int rs485_clean_rs(RS485INFO* pinfo) {
	int ret = 0;
	unsigned long irqflags;

	spin_lock_irqsave(&pinfo->rs_lock, irqflags);

	if (pinfo->use_mcua > 0) {
		ret = --(pinfo->use_mcua);
		if (pinfo->rs && pinfo->use_mcua == 0) {
			kfree(pinfo->rs);
			pinfo->rs = NULL;
		}
	}

	spin_unlock_irqrestore(&pinfo->rs_lock, irqflags);
	return ret;
}

inline void rs485_allow_interrupt_card(RS485INFO *pinfo) {
    u8 a = inb(pinfo->base +EPP_CONTROL);
#ifdef __M_BARRIER__
    rmb();
#endif
    outb( a | EPP_INT_ENABLE, pinfo->base +EPP_CONTROL );
#ifdef __M_BARRIER__
    wmb();
#endif
}

inline void rs485_stop_interrupt_card(RS485INFO *pinfo) {
    u8 a = inb(pinfo->base +EPP_CONTROL);
#ifdef __M_BARRIER__
    rmb();
#endif
    outb( a & ~EPP_INT_ENABLE, pinfo->base +EPP_CONTROL );
#ifdef __M_BARRIER__
    wmb();
#endif
}

inline void rs485_transmit_enable(RS485INFO *pinfo) {
    u8 a = inb(pinfo->base +EPP_CONTROL);
#ifdef __M_BARRIER__
    rmb();
#endif
    outb( a & ~ENABLE_BI_DIRECT, pinfo->base +EPP_CONTROL );
#ifdef __M_BARRIER__
    wmb();
#endif
}

inline void epp_data_flush( unsigned short base ) {
    unsigned int count = 0;
    while( ( (inb(base + EPP_STATUS)) & PARPORT_STATUS_ACK ) != 0 ) {
        inb( base + EPP_DATA );
        count++;
    }
    if( count ) PDEBUG( "lpt: flush %d bytes before return\n", count );
}

inline int is_data_ready( unsigned short base, unsigned int count ) {
    if( ( inb( base + EPP_STATUS) & PARPORT_STATUS_ACK ) == 0 ) {
	PDEBUG( "lpt: broken packet after %x symbol\n", count);
        return -1;
    }
    return 0;
}

int start_transmitter(RS485INFO *info) {

#ifndef KERN_2_6
    int ready=0;
    int resp_tmout=0;
#endif
    long remain_tm;
    const long timeout = HZ * RESP_TIMEOUT / 1000L / 1000L;     // JIFFIES
    data_ready = 0;                                             // not ready

    info->rs->usCountByteOut = 0;
    info->rs->status = PASSIVE_DATA;
////    outb( inb( info->base +EPP_CONTROL ) & ~ENABLE_BI_DIRECT, info->base +EPP_CONTROL );

    while( 1 ) {
        info->rs->usCountByteOut ++;                            // from 1-st, skipping address byte
        if( info->rs->usCountByteOut >= info->rs->outbuffer.size ) {
            break;
        }
        outb(info->rs->outbuffer.cache[info->rs->usCountByteOut], info->base +EPP_DATA);
#ifdef __M_BARRIER__
        wmb();
#endif
    }
    outb( info->rs->outbuffer.cache[0], info->base +EPP_ADDR ); // send address (end of the packet)
#ifdef __M_BARRIER__
    wmb();
#endif
////    outb( inb( info->base +EPP_CONTROL ) | ENABLE_BI_DIRECT, info->base +EPP_CONTROL );

#ifndef KERN_2_6
                                                                // waiting answer from module (in irq-handler)
    rs485_set_starttime();
    resp_tmout = 0;
    while( 1 ) {
        disable_irq(info.irq);                                  // protect race condition with handler
        ready = data_ready;
        enable_irq(info.irq);
        if(ready) break;
        if( rs485_timeout(RESP_TIMEOUT) ) {
            resp_tmout = 1;
            break;
        }
        udelay(RESP_WAIT);
    }
    if(resp_tmout) {
        PDEBUG( "lpt: transmit timeout!\n" );
        return -ERSNOCONVREPLY;
    } else {
        return(info->rs->inbuffer_pas.size);
    }
#else 
    remain_tm = wait_event_interruptible_timeout(tr_queue, data_ready != 0, timeout);
    if( signal_pending( current ) ) {
        return -ERESTARTSYS;
    }
    if( remain_tm == 0 ) {
        PDEBUG( "lpt: transmit timeout\n" );
        return -ERSNOCONVREPLY;
    }
    return(info->rs->inbuffer_pas.size);                        // signal or packet length
#endif
}


void receiv_data(RS485INFO * info) {

//  * -----------------------0x40------------------------------------
//  *        Receiver
//  *        if info->rs->status == PASSIVE_DATA, then it is answer to data request
//  *        passive-data format: [length][data...]  [signal]
//  *        active-data format: [address][length][data...]
//  * -----------------------------------------------------------

    static unsigned int enter_number;
    int ret = SUCCESS;
    struct _IN_BUFFER_ *current_inbuf;

#ifdef __DEBUG_TIMING__
    rs485_set_starttime();
#endif
    spin_lock(&info->rs_lock);
    
    if (info->use_mcua == 0 || info->rs == NULL) {
        goto end;
    }
    
    info->rs->usCountByteInp = 0;
    if( is_data_ready( info->base, info->rs->usCountByteInp ) < 0 ) 
        goto end;   // False enter

    if( info->rs->status == ACTIVE_DATA ) {
       
        current_inbuf = &(info->rs->inbuffer_act);
        enter_number++;                         // Address
        info->rs->tmp_byte = inb( info->base +EPP_DATA );
        if( info->rs->tmp_byte <  2*CACHE_MAXSIZE ) {
            current_inbuf->cache[info->rs->usCountByteInp++] = info->rs->tmp_byte;
            current_inbuf->adr  = info->rs->tmp_byte;
            PDEBUG( "lpt: address = %x enter = %d\n", info->rs->tmp_byte, enter_number);
        } else {
            PDEBUG( "lpt: signal in place of address = %x\n", info->rs->tmp_byte);
            epp_data_flush( info->base );
            goto end;
        }
        if( is_data_ready( info->base, info->rs->usCountByteInp ) < 0 ) goto end;
                                                // Packet lenght
        info->rs->tmp_byte = inb( info->base +EPP_DATA );
        if( info->rs->tmp_byte < PACKET_MAX_LENGTH+1 && info->rs->tmp_byte > 0 ) {
            current_inbuf->cache[info->rs->usCountByteInp++] = info->rs->tmp_byte;
            current_inbuf->size = info->rs->tmp_byte + 2;
        } else {
            PDEBUG( "lpt: incorrect length = %x\n", info->rs->tmp_byte);
            epp_data_flush( info->base );
            goto end;
        }
                                                // Data
        do {
	    if( is_data_ready( info->base, info->rs->usCountByteInp ) < 0 ) goto end;
            current_inbuf->cache[info->rs->usCountByteInp++] = inb( info->base +EPP_DATA );
        } while ( info->rs->usCountByteInp < current_inbuf->size );
                                                // Data in the FIFO
        if( ( ret = rs485_put_data( info ) ) != SUCCESS) {
            PDEBUG("lpt: return from rs485_put_data() = %d\n", ret);
        }
        epp_data_flush( info->base );
    }

    if( info->rs->status == PASSIVE_DATA ) {
        
        current_inbuf = &(info->rs->inbuffer_pas);
        current_inbuf->size = 0;
                                                // Signal from module
	info->rs->tmp_byte = inb( info->base + EPP_DATA );
	if( info->rs->tmp_byte > 2*CACHE_MAXSIZE ) {
	    current_inbuf->size = info->rs->tmp_byte;
            data_ready = 1;
            wake_up(&tr_queue);
            info->rs->status = ACTIVE_DATA;
            enter_number = 0;
            goto end;
        }
                                                 // Packet lenght
        if( info->rs->tmp_byte < PACKET_MAX_LENGTH+1 && info->rs->tmp_byte > 0 ) {
            current_inbuf->cache[info->rs->usCountByteInp++] = info->rs->tmp_byte;
            current_inbuf->size = info->rs->tmp_byte + 1;
        } else {
            PDEBUG( "lpt: incorrect length = %x\n", info->rs->tmp_byte);
            epp_data_flush( info->base );
            info->rs->status = ACTIVE_DATA;
            goto end;
        }
                                                // Data
        do {
	    if( is_data_ready( info->base, info->rs->usCountByteInp ) < 0 ) goto end;
            current_inbuf->cache[info->rs->usCountByteInp++] = inb( info->base +EPP_DATA );
        } while ( info->rs->usCountByteInp < current_inbuf->size );
        data_ready = 1;
        wake_up(&tr_queue);
        info->rs->status = ACTIVE_DATA;
    }

end:
    spin_unlock(&info->rs_lock);

#ifdef __DEBUG_TIMING__
    if( rs485_timeout( 200L ) ) RS_PRINT( "Long packet: > 200 usec!\n" );
#endif
    return;   
}

#ifdef __PP_PC__
#ifndef KERN_2_6_24
static void lpt_irq (int irq, void * private, struct pt_regs * unused) {
#else
static void lpt_irq (void * private) {
#endif
    RS485INFO* info = (RS485INFO *) private;
    receiv_data(info);
}
#endif

// -------------- FILE OPERARIONS --------------------------------------------


#define Fosc        14746
#define Minbaudrate 115
#define Maxbaudrate 1840

int ubrr_to_rate(int baudrate_code) {
    int a;
    a = baudrate_code + 1;
    if(a == 0) return 0;
    return( baudrate_code == 0x80 ? Fosc/8 : Fosc/16/a );
}

int rate_to_ubrr(int baudrate) {
    int baudrate_code;
    baudrate = ( baudrate < Minbaudrate )? Minbaudrate : baudrate;
    baudrate = ( baudrate > Maxbaudrate )? Maxbaudrate : baudrate;
    baudrate_code = baudrate > 1200 ? 0x80 : Fosc/16/baudrate - 1;
//VK new converter 1800-speed has code -1 ( 0xFF )
    return baudrate_code;
}

int rs485_set_baudrate(RS485INFO *pinfo) {
    int baudrate_code;
    baudrate_code = rate_to_ubrr(pinfo->baudrate);
    pinfo->baudrate = ubrr_to_rate(baudrate_code);
    PDEBUG("lpt: baudrate=%d\n",pinfo->baudrate);

    if( pinfo->rs->interface_type == CONV_VERS1 ) {
        pinfo->rs->outbuffer.size = 9;
        pinfo->rs->outbuffer.cache[0] = 0x1f;
        pinfo->rs->outbuffer.cache[1] = 0x07;
        pinfo->rs->outbuffer.cache[2] = 0x1f;
        pinfo->rs->outbuffer.cache[3] = 0x00;
        pinfo->rs->outbuffer.cache[4] = baudrate_code;
        pinfo->rs->outbuffer.cache[5] = 0x73;
        pinfo->rs->outbuffer.cache[6] = 0x00;
        pinfo->rs->outbuffer.cache[7] = 0x01;
        pinfo->rs->outbuffer.cache[8] = 0x6c;
    }
    if( pinfo->rs->interface_type == CONV_VERS2 ) {
        pinfo->rs->outbuffer.size = 3;
        pinfo->rs->outbuffer.cache[0] = 0x1f;
        pinfo->rs->outbuffer.cache[1] = 0x43;
        pinfo->rs->outbuffer.cache[2] = baudrate_code;
    }
    start_transmitter(pinfo);
    return SUCCESS;
}

int rs485_converter_reset(RS485INFO *pinfo) {
    if( pinfo->rs->interface_type == CONV_VERS2 ) {
        pinfo->rs->outbuffer.size = 2;
        pinfo->rs->outbuffer.cache[0] = 0x1f;
        pinfo->rs->outbuffer.cache[1] = 0x87;
    }
    start_transmitter(pinfo);
    return SUCCESS;
}

int rs485_test_device(RS485INFO *pinfo) {

    u8 iir = inb(pinfo->base+EPP_STATUS);
#ifdef __M_BARRIER__
    rmb();
#endif
    if( iir & PARPORT_STATUS_PAPEROUT ) {
        PDEBUG( "lpt: no converter\n");
        return -ERSNOINTERFACE;
    }

    if( iir & PARPORT_STATUS_SELECT ) {
        PDEBUG( "lpt: no power\n" );
        return -ERSSHUTDOWN;
    }

    if( iir & EPP_TIMEOUT ) {
        PDEBUG( "lpt: EPP timeout\n" );
        outb( EPP_TIMEOUT, pinfo->base +EPP_STATUS );
#ifdef __M_BARRIER__
        wmb();
#endif
    }
    return SUCCESS;
}

int rs485_reset( RS485INFO *pinfo ) {
    u8 i;
    int ret = SUCCESS;

    if (pinfo->base_hi) {
        i = inb(pinfo->base_hi+2);
        outb( 0x80 | (i & 0x1f), pinfo->base_hi+2);
        i = inb(pinfo->base_hi+2);
        if ((i & 0xe0) != 0x80) {
           PDEBUG("lpt: Parport is not in EPP mode: %x\n", i);
           return -ERSSHUTDOWN;
        }
    }
   
    ret = rs485_test_device( pinfo ); // Is converter connected and powered ? 
    if(ret != SUCCESS) return ret;

    outb( EPP_TIMEOUT, pinfo->base +EPP_STATUS );
#ifdef __M_BARRIER__
    wmb();
#endif
//    outb( 0, pinfo->base +EPP_CONTROL );                //reset
//#ifdef __M_BARRIER__
//    wmb();
//#endif
//    udelay(EPP_WAIT);    // 250 mksec
    outb( EPP_RESET, pinfo->base +EPP_CONTROL );
#ifdef __M_BARRIER__
    wmb();
#endif
    udelay(EPP_WAIT);    // 250 mksec
    outb( 0x0e, pinfo->base +EPP_DATA );
#ifdef __M_BARRIER__
    wmb();
#endif
    outb( 0x1f, pinfo->base +EPP_ADDR );
#ifdef __M_BARRIER__
    wmb();
#endif
////    outb( inb( pinfo->base +EPP_CONTROL ) | ENABLE_BI_DIRECT, pinfo->base +EPP_CONTROL );
    udelay(RESP_WAIT);    // 100 mksec
//    ii = inb( pinfo->base +EPP_STATUS );
    i = inb(pinfo->base+EPP_DATA);
#ifdef __M_BARRIER__
    rmb();
#endif
//    PDEBUG( "lpt: reset i = %x, csr=%x\n", i, ii );
      
    if( i == SignalACN ) {
        pinfo->rs->interface_type = CONV_VERS2;
        PDEBUG( "lpt: converter code vers.2\n" );
    } else  if (i == SignalACY) {
        pinfo->rs->interface_type = CONV_VERS1;
        PDEBUG( "lpt: converter code vers.1\n" );
    } else {
        PDEBUG( "lpt: converter doesn't respond\n" );
        return -ERSSHUTDOWN;    // na nekotorykh kompakh test_device ne srabatyvayet                       
    }
    udelay(EPP_WAIT);
//    rs485_converter_reset( pinfo );
    return ret;
}


void rs485_blnumber_reset(RS485INFO *info) {
    int i;
    for( i=0; i<MAX_TASK_DEVICE; i++ ) {
        info->rs->device[i].number = 0;
        info->rs->device[i].round_count_prev = -1;
    }
    info->rs->inbuffer_pas.adr = 0;
    info->rs->inbuffer_act.adr = 0;
}

int rs485_open(struct inode *inode, struct file *pfile) {
    int err = 0;
    unsigned int minor = iminor(inode);
    RS485INFO* info = &lpt_table[minor];

    if(minor >= PARPORT_MAX) return -ENXIO;
    if((lpt_table[minor].flags & EXIST_FL) == 0) return -ENXIO;

    if( down_interruptible(&info->sem) )
	return -ERESTARTSYS;
    
    if( info->use_mcua >= MAX_USE_STREAM ) return -ERSBUSY; // More than maximal shared users
    PDEBUG("lpt: open device %d, %d\n", imajor(inode), iminor(inode));

    err = rs485_init_rs(info);
    if (err == 0) { // First opening
        rs485_stop_interrupt_card(info);
        rs485_blnumber_reset(info);
        rs485_fifo_init(info);
        err = rs485_reset(info); 
        if (err != SUCCESS) {
            rs485_clean_rs(info);
            goto end;
        }
        rs485_allow_interrupt_card(info);
	rs485_converter_reset( info );
    } else if (err > 0) {
        err = 0;
    }

end:
    up(&info->sem);
    return err;
}

int rs485_release(struct inode *inode, struct file *pfile) {
    int err;
    unsigned int minor = iminor(pfile->f_dentry->d_inode);
    RS485INFO* info = &lpt_table[minor];

    if( down_interruptible(&info->sem ) )
	return -ERESTARTSYS;
    
    err = rs485_clean_rs(&lpt_table[minor]);
    if (err == 0) { // It was the last user
        rs485_stop_interrupt_card(info);
    }
    
    udelay(SEM_WAIT);
    up(&info->sem);
    return 0;
}

ssize_t rs485_write( struct file *fp, const char __user *buf, size_t size, loff_t *offset ) {

/*  Data write to port ( This isn't standard operation write )
 *  return:  < 0    - error code
 *           > 0x80 - signal
 *           <= 0x20 - length of the packet
 */    
    RS485INFO    *info = &(lpt_table[iminor(fp->f_dentry->d_inode)]);
    unsigned long rest = 0;
    int i, ret = 0;

    if( down_interruptible( &info->sem ) ) {
	return -ERESTARTSYS;
    }
    if( (size == 0) || (size > PACKET_MAX_LENGTH ) ) {
        PDEBUG( "lpt: write size=%zd is invalid!\n", size);
        ret =  -ERSFORMAT;
        goto OUT;
    }
    if( (ret=rs485_test_device(info)) < 0 ) goto OUT;
    rest = copy_from_user( info->rs->outbuffer.cache, buf, size );
    if(rest != 0) {
        PDEBUG("lpt: write copy_from_user rest=%ld\n", rest);
        ret = -EFAULT;
        goto OUT;
    }
    info->rs->outbuffer.size = size;
    for(i=0; i<MAX_HOP; i++) {
        ret = start_transmitter(info);
        if( (ret != SignalEER) && (ret != SignalNAK) ) break;
        msleep(4);
    }
    if( (ret > 0) && ( ret <= PACKET_MAX_LENGTH ) ) {
        rest = copy_to_user( (char*)buf, info->rs->inbuffer_pas.cache, ret);
        if( rest != 0 ) {
            PDEBUG( "lpt: write copy_to_user rest=%ld\n", rest );
            ret = -EFAULT;
            goto OUT;
        }       
    }
OUT:
    udelay(SEM_WAIT);
    up(&info->sem);
    return ret;
}

ssize_t rs485_read( struct file *fp, char __user *buf,
            size_t count, loff_t *offset ) {

    // Read data fifo
    
    char count0;
    RS485INFO    *info = &(lpt_table[iminor(fp->f_dentry->d_inode)]);
    int ret = 0;

    if( down_interruptible(&info->sem)) {        // 2.4 ???
	return -ERESTARTSYS;
    }

    ret = rs485_data_fifo_read( info, buf, 1 );
    if( ret > 0 ) {
        get_user(count0, buf);    // buf in User space!!!
        if( count0 < (count-1) ) count = count0;
        else count--;
        ret = rs485_data_fifo_read( info, buf+1, count );
    }
//        udelay(SEM_WAIT);
    up(&info->sem);
    return ret;
}


#ifdef KERN_2_6_36
long rs485_ioctl(struct file *fp,
            unsigned int cmd, unsigned long arg) {
#else
int rs485_ioctl(struct inode *inode, struct file *fp,
            unsigned int cmd, unsigned long arg) {
#endif

    int err = SUCCESS;
    RS485INFO *info;
    unsigned long rest = 0;
    
    if(_IOC_TYPE(cmd) != RS_IOC_MAGIC) {        // 2.4 ???
        PDEBUG("lpt: ioctl_error!\n");
        return -ENOTTY;
    }
    if(_IOC_NR(cmd) > RS_IOC_MAXNR) {
        PDEBUG("lpt: ioctl_error!\n");
        return -ENOTTY;
    }
#ifdef KERN_2_6
/*
    if(_IOC_DIR(cmd) & _IOC_READ) {
        acc = access_ok(VERIFY_WRITE, (void __user *)arg, _IOC_SIZE(cmd));
    } else {
        acc = access_ok(VERIFY_READ, (void __user *)arg, _IOC_SIZE(cmd));
    }

    if(!acc) {
        PDEBUG("ioctl_error!\n");
        return -EFAULT;
    }
*/
#endif
    info = &(lpt_table[iminor(fp->f_dentry->d_inode)]);

    switch(cmd) {
        case RS_RESET:
            if( down_interruptible(&info->sem)) {
                up( &info->sem );
                return -EINTR;
            }
            rs485_stop_interrupt_card( info );
            err = rs485_reset( info );
            rs485_fifo_flush( info );
            if( err == SUCCESS ) {
                rs485_allow_interrupt_card( info );
            }
            udelay( SEM_WAIT );
            up( &info->sem );
            
            return err;

        case RS_FIFO_FLUSH:
            if( down_interruptible(&info->sem)) {
                up(&info->sem);
                return -EINTR;
            }
            rs485_fifo_flush(info);
            udelay(SEM_WAIT);
            up(&info->sem);
            return SUCCESS;

        case RS_SET_BAUDRATE:
            if( down_interruptible(&info->sem)) {
                up(&info->sem);
                return -EINTR;
            }
            err = rs485_test_device(info);
            if(err >= 0 ) {
                info->baudrate = (int)arg;
                err = rs485_set_baudrate(info);
            }
            udelay(SEM_WAIT);
            up(&info->sem);
            return err;

        case RS_GET_BAUDRATE:
            return info->baudrate;

        case RS_BLNUMBER_RESET:
            rs485_blnumber_reset(info);
            return SUCCESS;

        case RS_GET_VERS:
            rest = copy_to_user((char*)arg, RS485_UTS_RELEASE, sizeof(RS485_UTS_RELEASE));
            if(rest != 0) {
                PDEBUG("lpt: ioctl get_vers, copy_to_user rest=%ld\n", rest);
                return -EFAULT;
            }
            return RS485_VERSION_CODE;

        case RS_GET_DATAFIFO_SIZE:
            return SIZE_STREAM_FIFO;

        default: break;
    }
    return -ERSNOIOCTL;
}

struct file_operations rs485_fops = {
    .owner        = THIS_MODULE,
    .open         = rs485_open,
    .release      = rs485_release,
    .read         = rs485_read,
    .write        = rs485_write,
#ifdef KERN_2_6_36
    .unlocked_ioctl = rs485_ioctl,
#else
    .ioctl        = rs485_ioctl,
#endif
};


// ------------------------------- INIT / EXIT ----------------------------

#ifdef __PP_PC__
#ifdef KERN_2_6_18
static struct class *lptdev_class;
#else
static struct class_simple *lptdev_class;
#endif

static unsigned int lpt_count = 0;
static void lpt_attach(struct parport *port) {
    struct pardevice * pdev = NULL;
    PDEBUG("lpt: port %s attached\n", port->name);
//    pdev = parport_register_device (port, ADAPTER_ID, NULL,
//                                NULL, lpt_irq, PARPORT_DEV_EXCL, &lpt_table[lpt_count]);
    pdev = parport_register_device (port, ADAPTER_ID, NULL,
                                NULL, lpt_irq, 0, &lpt_table[lpt_count]);
    if (!pdev) {
        printk (KERN_WARNING "%s: failed to register device!\n", ADAPTER_ID);
        return;// -ENXIO;
    }
//    parport_put_port (port);
    if(parport_claim(pdev)) {
        PDEBUG("lpt: parport is not claimed\n");
        parport_unregister_device(pdev);
        return; //-EBUSY
    }
    lpt_table[lpt_count].pdev = pdev;
    lpt_table[lpt_count].flags |= EXIST_FL;
    lpt_table[lpt_count].base = pdev->port->base;
    lpt_table[lpt_count].base_hi = pdev->port->base_hi;
    lpt_table[lpt_count].irq = pdev->port->irq;
    printk (KERN_DEBUG "%s: registered pardevice %d\n", ADAPTER_ID, lpt_count);

#ifdef KERN_2_6_27
    device_create(lptdev_class, port->dev, MKDEV(major, lpt_count), NULL,
                    "rs485lpt%d", lpt_count);
#else
 #ifdef KERN_2_6_24
    device_create(lptdev_class, port->dev, MKDEV(major, lpt_count),
                    "rs485lpt%d", lpt_count);
 #else
  #ifdef KERN_2_6_18
    class_device_create(lptdev_class, NULL, MKDEV(major, lpt_count), NULL,
                    "rs485lpt%d", lpt_count);
  #else
    class_simple_device_add(lptdev_class, MKDEV(major, lpt_count), NULL,
                    "rs485lpt%d", lpt_count);
    devfs_mk_cdev( MKDEV(major, lpt_count), S_IFCHR | S_IRUGO | S_IWUGO,
                     "mcua%d", lpt_count);
  #endif
 #endif
#endif
    printk(KERN_INFO "rs485lpt%d: using %s (%s).\n", port->number, port->name,
        (port->irq == PARPORT_IRQ_NONE)?"polling":"interrupt-driven");
    PDEBUG("lpt: attach io=0x%lX io_hi=0x%lX irq=%d\n", port->base, port->base_hi, port->irq);
    lpt_count++;
}

static void lpt_detach(struct parport *port) {
    PDEBUG("lpt: %s  detached\n", port->name);
}

static struct parport_driver lpt_driver = {
    .name           = ADAPTER_ID,
    .attach         = lpt_attach,
    .detach         = lpt_detach,
};

#endif

#ifndef __PP_PC__
static void rs485_release_resource(void) {
    int i;
    for(i=0; i<PARPORT_MAX; i++) {
        if(lpt_table[i].flags & EXIST_FL) {
            release_region(lpt_table[i].base, ADDRESS_REGION_SIZE);
            rs485_clean_irq(&lpt_table[i]);
//            PDEBUG("irq=%d cleaned\n", lpt_table[i].irq);
//            PDEBUG("region %x released\n", lpt_table[i].base);
            lpt_table[i].flags &= ~EXIST_FL;
        }
    }
    return;
}

void __init rs485_request_resource(void) { // IO ports & irqs initialization
    int i;
    for(i=0; i<PARPORT_MAX; i++) {
#ifndef KERN_2_6
        if(check_region(lpt_table[i].base, ADDRESS_REGION_SIZE) != 0 ) {
//            RS_PRINT( KERN_CRIT "%s: I/O port 0x%08lX in use.\n",
//                        ADAPTER_ID, (unsigned long) lpt_table[i].base);
            continue;
        }
        request_region(lpt_table[i].base, ADDRESS_REGION_SIZE, ADAPTER_ID);
#else
        if(!request_region(lpt_table[i].base, ADDRESS_REGION_SIZE, ADAPTER_ID)) {
//            RS_PRINT( KERN_CRIT "%s: I/O port 0x%08lX in use.\n",
//                            ADAPTER_ID, (unsigned long) lpt_table[i].base);
            continue;
        }
#endif
        PDEBUG("lpt: port=%d, base=0x%08lX requested\n", i, (unsigned long) lpt_table[i].base);
        if(rs485_request_irq(&lpt_table[i])) {
            release_region(lpt_table[i].base, ADDRESS_REGION_SIZE);
            printk("region 0x%08lX released\n", (unsigned long)lpt_table[i].base);
            continue;
        }
        PDEBUG("lpt: port=%d, irq=%d requested\n", i, lpt_table[i].irq);
        lpt_table[i].flags |= EXIST_FL;
    }
}
#endif

MODULE_AUTHOR("Andre Zaitsev, Olga Voziakova (ovoz@sai.msu.ru)");
MODULE_DESCRIPTION("rs485 lpt module");
#ifdef MODULE_LICENSE
    MODULE_LICENSE("GPL");
#endif

#ifndef KERN_2_6_18
static unsigned int tmpvar=0;
#else
#define tmpvar NULL
#endif
#ifndef __PP_PC__
MODULE_PARM_DESC(io, "Base I/O address (hex), default io1=_IO_LPT1, io2=_IO_LPT2");
module_param_array(io, int, tmpvar, 0);
MODULE_PARM_DESC(irq, "IRQ line,default irq1=_IRQ_LPT1, irq2=_IRQ_LPT2");
module_param_array(irq, int, tmpvar, 0);
#endif
MODULE_PARM_DESC(baudrate, "BAUD RATE UART port, default baudrate1=_BAUDRATE_1, baudrate2=_BAUDRATE_2");
module_param_array(baudrate, int, tmpvar, 0);

#ifndef KERN_2_6
EXPORT_NO_SYMBOLS;
#endif

int __init init_module(void) {
    int err = 0;
    unsigned int i;
    char version[]=RS485_UTS_RELEASE;

    for(i=0; i<PARPORT_MAX; i++) {
        lpt_table[i].rs = NULL;
        spin_lock_init(&lpt_table[i].rs_lock);
        lpt_table[i].use_mcua = 0;
        lpt_table[i].flags = 0;
        init_MUTEX(&lpt_table[i].sem);
        lpt_table[i].baudrate = baudrate[i];
#ifndef __PP_PC__
        lpt_table[i].base = io[i];
        lpt_table[i].base_hi = io[i] + 0x400;
        lpt_table[i].irq = irq[i];
#else
        lpt_table[i].base = 0;
        lpt_table[i].base_hi = 0;
        lpt_table[i].irq = -1;
        lpt_table[i].pdev = NULL;
#endif
//        PDEBUG("init port %d: baudrate=%d base=%x irq=%d\n",
//            i+1, lpt_table[i].baudrate, lpt_table[i].base, lpt_table[i].irq);
    }

#ifndef __PP_PC__
    rs485_request_resource();
#endif

    if ( (major=register_chrdev (0, ADAPTER_ID, &rs485_fops)) <= 0) {
        printk ("rs485lpt: unable to get major\n");
        err = -EIO;
        goto out_chrdev;
    }
#ifndef __PP_PC__
    rs485_activate_parport();
#else

#ifdef KERN_2_6_18
    lptdev_class = class_create(THIS_MODULE, ADAPTER_ID);
#else
    devfs_mk_dir("mcua");
    lptdev_class = class_simple_create(THIS_MODULE, ADAPTER_ID);
    if (IS_ERR(lptdev_class)) {
        err = PTR_ERR(lptdev_class);
        goto out_class ;
    }
#endif
    if (parport_register_driver(&lpt_driver)) {
        printk ("rs485lpt: unable to register with parport\n");
        err = -EIO;
        goto out_register_driver;
    }
#endif  // __PP_PC__
    RS_PRINT(KERN_INFO "%s module, version %s\n",
                ADAPTER_ID, version);
    for(i=0; i<PARPORT_MAX; i++) {
        if(lpt_table[i].flags & EXIST_FL) {
            RS_PRINT(KERN_INFO "%s: install card on the base 0x%X (0x%X) and irq=%d\n",
            ADAPTER_ID, lpt_table[i].base, lpt_table[i].base_hi, lpt_table[i].irq);
        }
    }
    return 0;

#ifdef __PP_PC__
    parport_unregister_driver(&lpt_driver);
out_register_driver:
 #ifdef KERN_2_6_18
    class_destroy(lptdev_class);
 #else
    class_simple_destroy(lptdev_class);
out_class:
    devfs_remove("mcua");
 #endif
#else
    rs485_stop_parport();
#endif
    unregister_chrdev(major, ADAPTER_ID);
out_chrdev:
#ifndef __PP_PC__
    rs485_release_resource();
#endif
    return err;
}

void __exit cleanup_module(void){
#ifndef __PP_PC__
    rs485_stop_parport();
    rs485_release_resource();
#else
    int i;
    parport_unregister_driver(&lpt_driver);

    for(i=0; i<PARPORT_MAX; i++) {
        if(!(lpt_table[i].flags & EXIST_FL)) continue;
        parport_release(lpt_table[i].pdev);
        parport_unregister_device(lpt_table[i].pdev);
        printk ("%s%x: unregistered pardevice %d\n",ADAPTER_ID, i, i);
#ifdef KERN_2_6_24
        device_destroy(lptdev_class, MKDEV(major, i));
#else
#ifdef KERN_2_6_18
        class_device_destroy(lptdev_class, MKDEV(major, i));
#else
        devfs_remove("mcua/%d",i);
        class_simple_device_remove(MKDEV(major, i));
#endif
#endif
//        PDEBUG("class device minor=%d destroyed\n",i);
        lpt_table[i].flags &= ~EXIST_FL;
    }
#ifdef KERN_2_6_18
    class_destroy(lptdev_class);
#else
    devfs_remove("mcua");
    class_simple_destroy(lptdev_class);
#endif
//    PDEBUG("class destroyed\n");
#endif // PP_PC
    unregister_chrdev (major, ADAPTER_ID);
    return;
}
